#include <string.h>
#include <malloc.h>
#include <usb.h>

#include "hw.h"
#include "nsapp.h"
#include "fnt5x5.h"

#define TSTEP 50000

int hw_write_ctrl(hwctx_t *ctx, unsigned val)
{
	return hw_write_reg(ctx, 0x1FFF, val);
}

int hw_xor_byte(hwctx_t *ctx, int idx, unsigned char val)
{
	int i, ones[8], n=0;
	idx*=8;
	for(i=0;i<8;i++)
		if(val&(1<<i))
			ones[n++]=i;
	if(!n)
		return 0;
	while(n--)
		if(n>0) {
			if(hw_write_reg(ctx, 0x1C01, (ones[n]+idx)|((ones[n-1]+idx)<<9)))
				return 1;
			n--;
		} else
			if(hw_write_reg(ctx, 0x1C00, (ones[n]+idx)))
				return 1;
	return 0;
}

int hw_chip_mask(hwctx_t *ctx, unsigned mask)
{
	return hw_write_word(ctx, 0x80010000|mask);		// "set chip mask"
}

int hw_irq_mask(hwctx_t *ctx, unsigned mask)
{
	return hw_write_word(ctx, 0x80030000|mask);		// "set irq mask"
}

int hw_write_reg(hwctx_t *ctx, unsigned reg, unsigned val)
{
	return hw_write_word(ctx, ((reg&0x1FFF)<<18)|(val&0x3FFFF));	// "write register"
}

int hw_read_reg(hwctx_t *ctx, unsigned reg, unsigned count)
{
	return hw_write_word(ctx, 0x80000000|((reg&0x1FFF)<<18)|((count-1)&0xF));	// "read registers"
}

int hw_led_char(hwctx_t *ctx, int pos, unsigned char ch)
{
	int i, base = (pos<8)?0x80020100:0x80020200;		// "LED, display 1" or "LED, display 2"
	if(hw_write_word(ctx, base|0xA0|(pos&7)))		// "set position"
		return 1;
	for(i=0;i<5;i++)
		if(hw_write_word(ctx, base|font5x5[ch*5+i]))	// "write character"
			return 1;
	return 0;
}
int hw_led_line(hwctx_t *ctx, char *str)
{
	int i;
	for(i=0;str[i] && i<16;i++)
		if(hw_led_char(ctx, i, str[i]))
			return 1;
	for(;i<16;i++)
		if(hw_led_char(ctx, i, ' '))
			return 1;
	return 0;
}

int hw_write_word(hwctx_t *ctx, unsigned val)
{
	if(ftdi_write_data(&ctx->ftdi, (void *)&val, 4)<4)
		retval_message(1, MSG_FAIL, "Write failed");
	return 0;
}

static int hw_read_byte(hwctx_t *ctx, unsigned char *val, int timeout)
{
	int res;
	while(1) {
		if((res=ftdi_read_data(&ctx->ftdi, val, 1))==1)
			return 0;
		if(res<0)
			retval_message(1, MSG_FAIL, "Read failed");
		usleep(TSTEP);
		timeout -= TSTEP;
		if(timeout<0)
			return -1;
	}
}

int hw_read_word(hwctx_t *ctx, unsigned *ret, int timeout)
{
	int res;
	unsigned char val;
	*ret = 0xCDCDCDCD;
	while(ctx->read_pos < 24) {
		res = hw_read_byte(ctx, &val, timeout);
		if(res)
			return res;
		ctx->read_buf |= ((unsigned)val)<<ctx->read_pos;
		ctx->read_pos += 8;
	}
	*ret = ctx->read_buf;
	ctx->read_buf = 0;
	ctx->read_pos = 0;
	return 0;
}

hwctx_t *hw_open_board(char *serial)
{
	hwctx_t *ctx = calloc(1, sizeof(hwctx_t));
	ftdi_init(&ctx->ftdi);
	ftdi_set_interface(&ctx->ftdi, INTERFACE_B);
	if(ftdi_usb_open_desc(&ctx->ftdi, 0x8482, 0x1002, NULL, serial))
		retval_message(NULL, MSG_FAIL, "Opening board %s failed", serial);
	ftdi_usb_purge_buffers(&ctx->ftdi);
	usleep(100000);
	return ctx;
}

void hw_close_board(hwctx_t *ctx)
{
	ftdi_deinit(&ctx->ftdi);
}

char **hw_find_boards(void)
{
	struct usb_bus *bus;
	struct usb_device *dev;
	struct usb_dev_handle *inst;
	char str[256];
	int ndev = 0, strp = 0, wptr = 0;
	char **tab, *sym;

	usb_init();
	if(usb_find_busses() < 0)
		retval_message(NULL, MSG_FAIL, "No USB busses found");
	if(usb_find_devices() < 0)
		retval_message(NULL, MSG_FAIL, "No USB devices found");
	
	for(bus=usb_busses; bus; bus=bus->next)
		for(dev=bus->devices; dev; dev=dev->next)
			if(dev->descriptor.idVendor == USB_VENDOR && dev->descriptor.idProduct == USB_DEVICE) {
				if(!(inst=usb_open(dev))) {
					message(MSG_WARN, "Cannot open USB device %d:%d", dev->bus->location, dev->devnum);
					continue;
				}
				if(usb_get_string_simple(inst, dev->descriptor.iSerialNumber, str, sizeof(str)) <= 0) {
					usb_close(inst);
					message(MSG_WARN, "Cannot get serial for USB device %d:%d", dev->bus->location, dev->devnum);
					continue;
				}
				ndev ++;
				strp += strlen(str)+1;
				usb_close(inst);
			}

	if(!ndev)
		retval_message(NULL, MSG_FAIL, "No devices found");

	tab = calloc((ndev+1)*sizeof(char *)+strp, 1);
	sym = (char *)(tab+(ndev+1));
	ndev = 0;

	for(bus=usb_busses; bus; bus=bus->next)
		for(dev=bus->devices; dev; dev=dev->next)
			if(dev->descriptor.idVendor == USB_VENDOR && dev->descriptor.idProduct == USB_DEVICE) {
				if(!(inst=usb_open(dev))) {
					message(MSG_WARN, "Cannot open USB device %d:%d", dev->bus->location, dev->devnum);
					continue;
				}
				if(usb_get_string_simple(inst, dev->descriptor.iSerialNumber, str, sizeof(str)) <= 0) {
					usb_close(inst);
					message(MSG_WARN, "Cannot get serial for USB device %d:%d", dev->bus->location, dev->devnum);
					continue;
				}
				wptr += strlen(str)+1;
				if(wptr > strp) {
					free(tab);
					retval_message(NULL, MSG_FAIL, "USB devices changed while enumerating");
				}
				tab[ndev] = sym;
				strcpy(sym, str);
				ndev ++;
				sym += strlen(str)+1;
				usb_close(inst);
			}

	message(MSG_INFO, "Found %d boards", ndev);
	return tab;
}
